import logging
from datetime import datetime

from .providers import providers, lights, sensors
from ..managers import Manager
from ...protocols import dali
from ...protocols.dali.device import events
from ...protocols.dali.driver.dali_interface import DaliCanAdapter, BrokerAdapter
from ...tools.settings import config, LOCAL_ADAPTER, LEVEL_SYNC_DELAY

DALI1 = 1
DALI2 = 2


class Dali(Manager):
    DTR0 = 0
    DTR1 = 1
    DTR2 = 2

    level_sync_delay = 1

    def __init__(self, p_id, m_id):
        super().__init__(p_id, m_id)
        self._groups = [None, [[] for _ in range(0, 16)], [[] for _ in range(0, 32)]]
        self._broadcast = [None]*3
        self.group_providers = (None, [None for _ in range(0, 16)], [None for _ in range(0, 32)])
        self.dtr_values = [None, [None]*3, [None]*3]

    def _sorted_info(self):
        def s(_provider):
            if _provider.dali_type == DALI2:
                return _provider.dali_type, 10-_provider.instance._value, _provider.id
            else:
                return _provider.dali_type, _provider.id

        yield from sorted(self.providers.values(), key=s)

    def set_interface(self):
        self.send_interface = DaliCanAdapter() if config.get(LOCAL_ADAPTER, False) else BrokerAdapter(self.mqttc)
        logging.info('Bus interface is "%s"' % self._send_interface.__class__.__name__)

    def update_config(self):
        super(Dali, self).update_config()
        if self._send_interface and isinstance(self._send_interface, DaliCanAdapter) != config.get(LOCAL_ADAPTER, False):
            self.set_interface()

        self.level_sync_delay = config.get(LEVEL_SYNC_DELAY, self.level_sync_delay)
        from spread_core.bam.dali.tuner import TunerLimited
        TunerLimited.update_config()

    def add_provider(self, data):
        super(Dali, self).add_provider(data)
        provider = self.providers[data['id']]
        if provider.is_device:
            addr = provider.binding.addr.address
            self.devices.setdefault(addr, [])
            self.devices[addr].append(provider)
        elif provider.is_group:
            self.group_providers[provider.dali_type][provider.binding.addr.group] = provider
        elif provider.is_broadcast:
            self._broadcast[provider.dali_type] = provider

    def broadcast(self, dali_type):
        return self._broadcast[dali_type]

    def add_to_group(self, provider, group_id):
        if provider not in self._groups[provider.dali_type][group_id]:
            self._groups[provider.dali_type][group_id].append(provider)

    def remove_from_group(self, provider, group_id):
        if provider in self._groups[provider.dali_type][group_id]:
            self._groups[provider.dali_type][group_id].remove(provider)

    def get_group_size(self, dali_type, group_id):
        return len(self._groups[dali_type][group_id])

    def get_group_members(self, dali_type, group_id):
        yield from self._groups[dali_type][group_id]

    def get_all_devices(self, dali_type):
        for _provider in self.providers.values():
            if _provider.is_device and _provider.dali_type == dali_type:
                yield _provider

    def on_external_command(self, command, response, *args):
        if isinstance(command, dali.gear.general.DTR0):
            self.dtr_values[DALI1][0] = command.param
        elif isinstance(command, dali.gear.general.DTR1):
            self.dtr_values[DALI1][1] = command.param
        elif isinstance(command, dali.gear.general.DTR2):
            self.dtr_values[DALI1][2] = command.param
        elif isinstance(command, dali.device.general.DTR0):
            self.dtr_values[DALI2][0] = command._opcode
        elif isinstance(command, dali.device.general.DTR1):
            self.dtr_values[DALI2][1] = command._opcode
        elif isinstance(command, dali.device.general.DTR2):
            self.dtr_values[DALI2][2] = command._opcode
        elif isinstance(command, dali.device.general.DTR1DTR0):
            self.dtr_values[DALI2][0] = command._opcode
            self.dtr_values[DALI2][1] = command._instance
        elif isinstance(command, dali.device.general.DTR2DTR1):
            self.dtr_values[DALI2][1] = command._opcode
            self.dtr_values[DALI2][2] = command._instance
        elif isinstance(command, (dali.gear.general._StandardCommand, dali.gear.general.DAPC)):
            if isinstance(command.destination, dali.address.Short):
                devices = self.providers_by_address(DALI1, command.destination.address)
            elif isinstance(command.destination, dali.address.Group):
                group = self.group_providers[DALI1][command.destination.group]
                if group:
                    devices = (group,)
                else:
                    devices = self.get_group_members(DALI1, command.destination.group)
            elif isinstance(command.destination, dali.address.Broadcast):
                if self._broadcast[DALI1]:
                    devices = (self._broadcast[DALI1],)
                else:
                    devices = self.get_all_devices(DALI1)
            else:
                devices = ()

            for _provider in devices:
                if _provider is None: continue
                if not isinstance(_provider, lights.DaliLight): continue

                if command.is_query:
                    [_provider.on_request_answer(funit_type, response, None, *args) for funit_type in command.funit_type]
                else:
                    if _provider.is_valid:
                        _provider.on_command_sended(command)

        elif isinstance(command, dali.device.general._DeviceCommand):
            # DALI_2
            if isinstance(command, dali.device.general._SpecialDeviceCommand):
                pass
            else:
                if isinstance(command.destination, dali.address.Short):
                    devices = self.providers_by_address(DALI2, command.destination.address)
                elif isinstance(command.destination, dali.address.Broadcast):
                    devices = self.get_all_devices(DALI2)
                elif isinstance(command.destination, dali.address.Group):
                    group = self.group_providers[DALI2][command.destination.group]
                    if group:
                        devices = (group,)
                    else:
                        devices = self.get_group_members(DALI2, command.destination.group)
                else:
                    devices = ()

                for _provider in devices:
                    if isinstance(command, dali.device.general._StandardInstanceCommand):
                        if not isinstance(command.instance, (dali.address.InstanceGroup, dali.address.InstanceBroadcast)):
                            if not _provider.compare_instance(command.instance):
                                continue

                    if command.is_query:
                        [_provider.on_request_answer(funit_type, response, None, *args) for funit_type in command.funit_type]
                    elif command.is_config:
                        if _provider.is_valid:
                            _provider.on_command_sended(command)

    def on_event(self, event_data):
        event = events.Event.of(event_data)

        if event.scheme_id == dali.device.events.SCHEME_ADDR_INST_TYPE or event.scheme_id == dali.device.events.SCHEME_ADDR_INST_NUMBER:
            for _provider in self.devices.get(event.short_addr.address, ()):
                if isinstance(_provider, sensors.DaliSensor) and _provider.compare_instance(event.instance_type or event.instance_number):
                    event.data = _provider.type_id
                    logging.debug(f'<- {event}')
                    _provider.on_event(event)
                    break
            else:
                logging.debug(f'<- {event}')
