import logging

import spread_knx as knx
from spread_core.bam import Funit, Provider, MultiCommand


class KnxProvider(Provider):
    def_state = None
    address_class = None

    def __init__(self, data):
        super(KnxProvider, self).__init__(data)
        self._address = self.address_class(data.get('attributes'))
        self._status = self.SUSPICIOUS

    @property
    def address(self):
        return self._address

    def parse_bindings(self, data):
        pass

    def on_survey(self):
        yield MultiCommand(self.check_valid)

    def request(self, cmd, *args, sig=None, state_address=None, **kwargs):
        for response in self._manager.send_interface.send(cmd, state_address):
            self.on_request_answer(response, *args, sig=sig)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None, **kwargs):
        logging.warning('Unknown response %s' % response)

    def set_valid(self, validity):
        if validity == self._status == self.SUSPICIOUS:
            validity = self.INVALID

        super(KnxProvider, self).set_valid(validity)

        if validity != self.SUSPICIOUS:
            for funit_type in self.values.copy():
                if funit_type not in (PositionState, PositionControl, StopControl):
                    self.on_update(funit_type, None if validity == self.INVALID else self.get_value(funit_type))

    def check_valid(self):
        raise NotImplementedError()

    def get_info(self):
        pass

    def prs2raw(self, p):
        if p == 0:
            return 0
        return min(max(1, round(255 * p / 100)), 255)

    def raw2prs(self, r):
        return None if r is None else min(100, round(100 * r / 255))

    def SetDiscovery(self, value, sig=None):
        pass

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        raise NotImplemented()

    def GetBrightnessLevel(self, sig=None):
        raise NotImplemented()


class KnxLights(KnxProvider):
    address_class = knx.address.LightAddress

    @property
    def address(self) -> address_class:
        return super(KnxLights, self).address

    def repair_on_fail(self) -> tuple:
        return (*super(KnxProvider, self).repair_on_fail(), BrightnessLevel,)

    def check_valid(self):
        self.GetBrightnessLevel()

        if self.get_value(BrightnessLevel) is None:
            self.set_valid(self.SUSPICIOUS)

    def get_bindings(self):
        self.GetState()
        self.GetControl()

    def GetState(self, sig=None):
        self.on_update(State, self.address.state, sig=sig)

    def GetControl(self, sig=None):
        self.on_update(Control, self.address.control, sig=sig)


class KnxDimmer(KnxLights):
    def_state = knx.commands.Level

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Level):
            self.on_update(BrightnessLevel, self.raw2prs(response.value), sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, preliminary=True, sig=sig)
        self.request(knx.commands.Level.set(self.address.control, self.prs2raw(value)), sig=sig)

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Level.get(self.address.state), state_address=self.address.state, sig=sig)


class KnxRelay(KnxLights):
    def_state = knx.commands.Switching

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Switching):
            value = response.value
            if value is not None:
                value = 100 if value else 0
            self.on_update(BrightnessLevel, value, sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, preliminary=True, sig=sig)
        self.request(knx.commands.Switching.set(self.address.control, value > 0), sig=sig)

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Switching.get(self.address.state), state_address=self.address.state, sig=sig)


class KnxMotor(KnxProvider):
    address_class = knx.address.MotorAddress
    def_state = knx.commands.Level

    @property
    def address(self) -> address_class:
        return super(KnxMotor, self).address

    def repair_on_fail(self) -> tuple:
        return (*super(KnxProvider, self).repair_on_fail(), PositionLevel,)

    def check_valid(self):
        self.GetLevel()

        if self.get_value(PositionLevel) is None:
            self.set_valid(self.SUSPICIOUS)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Level):
            self.on_update(PositionLevel, self.raw2prs(response.value), sig=sig)

    def get_bindings(self):
        self.GetPositionState()
        self.GetPositionControl()
        self.GetStopControl()

    def GetPositionState(self, sig=None):
        self.on_update(PositionState, self.address.state, sig=sig)

    def GetPositionControl(self, sig=None):
        self.on_update(PositionControl, self.address.control, sig=sig)

    def GetStopControl(self, sig=None):
        self.on_update(StopControl, self.address.stop, sig=sig)

    def SetLevel(self, value, sig=None, **kwargs):
        if value is None:
            self.request(knx.commands.Step.set(self.address.stop, 1), sig=sig)
            # Значение возвращается само по завершении дижения
            # self.GetLevel()
        else:
            level = self.prs2raw(value)
            has_level = self.get_value(PositionLevel)

            self.on_update(PositionLevel, value, preliminary=True, sig=sig)

            if level == has_level:
                self.GetLevel(sig=sig)
            else:
                self.request(knx.commands.Level.set(self.address.control, level), sig=sig)

    def GetLevel(self, sig=None):
        self.request(knx.commands.Level.get(self.address.state), state_address=self.address.state, sig=sig)


class KNXFunit(Funit):
    entity_type = KnxProvider
    easy_get = False
    get_after_set = False
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None

    @property
    def entity(self) -> KnxProvider:
        return super(KNXFunit, self).entity

    def set(self):
        if self.set_cmd:
            self.entity.manager.add_command(self.entity, self.set_cmd.set(self.entity.address.control, self.set_value))
        elif self.set_func:
            self.entity.manager.add_command(
                self.entity,
                self._apply_func(self.entity, self.set_func.__name__)
            )

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        elif self.get_cmd:
            self.entity.manager.add_command(self.entity, self.get_cmd.get(self.entity.address.state))
        elif self.get_func:
            self.entity.manager.add_command(
                self.entity,
                self._apply_func(self.entity, self.get_func.__name__)
            )


KnxProvider.funit_class = KNXFunit


class KnxLightFunit(KNXFunit):
    pass


class State(KnxLightFunit):
    get_func = KnxLights.GetState


class Control(KnxLightFunit):
    get_func = KnxLights.GetControl


class BrightnessLevel(KnxLightFunit):
    set_func = KnxProvider.SetBrightnessLevel
    get_func = KnxProvider.GetBrightnessLevel
    value_range = range(0, 101)


class GroupLevel(BrightnessLevel):
    pass


class KnxMotorFunit(KNXFunit):
    pass


class PositionState(KnxMotorFunit):
    get_func = KnxMotor.GetPositionState


class PositionControl(KnxMotorFunit):
    get_func = KnxMotor.GetPositionControl


class StopControl(KnxMotorFunit):
    get_func = KnxMotor.GetStopControl


class PositionLevel(KnxMotorFunit):
    set_func = KnxMotor.SetLevel
    get_func = KnxMotor.GetLevel
    value_range = range(0, 101)
