import json
import logging

import spread_core.bam as bam
import spread_core.mqtt as mqtt
from spread_core.errors.project_errors import ProjectError, ProviderNotPresent, EngineryNotPresent, ClassifierError, \
    CommandError
from spread_core.tools.manager_interface import ManagerOfBroker

PROVIDERS = 'providers'
ID = 'id'
TYPE = 'type'
MANAGER_ID = 'managerID'
RECIPE = 'recipe'
INGREDIENTS = 'ingredients'


def log_command(fn):
    def log(instance, topic, variable):
        logging.debug(
            f'COMMAND({variable.session_id if variable.session_id else ""}): {topic.__class__.__name__} {topic.entity_address.type}({topic.entity_address.id})'
            f'.{topic.entity_address.funit_type}' +
            (f'={variable.value}' if isinstance(topic, mqtt.spread.topic.Set) else '') +
            (f' params: {variable.params}' if variable.params else '')
        )
        fn(instance, topic, variable)

    return log


class Equipment(ManagerOfBroker):
    STATE_TOPICS = None

    def __init__(self, project_id, mqttc):
        super(Equipment, self).__init__(mqttc)
        self.id = project_id
        self._managers = dict()
        self._subgineries = dict()
        self._engineries = {}
        self.loaded = False

    @classmethod
    def enginery_types(cls): return []

    @classmethod
    def subginery_types(cls): return []

    def stack_for_reload(self) -> tuple:
        """Список коллекций объектов, для которых необходимо переопубликовать значения"""
        yield from self._engineries
        yield from self._subgineries

    def start(self):
        if self.loaded:
            for s in self.stack_for_reload():
                for e in s.values():
                    values = {**e.values}
                    e.values = {}
                    for funit_type, value in values.items():
                        e.publish_state(funit_type=funit_type, value=value)

            self.exploit_topics(self.subscribe)
        else:
            self.subscribe(mqtt.project.TopicProject.managers(self.id))

    def on_project(self, topic, data):
        self.unsubscribe(topic)
        data = json.loads(data)
        if mqtt.project.TopicProject.MANAGERS in data:
            self.on_project_managers(data[mqtt.project.TopicProject.MANAGERS])
        elif mqtt.project.TopicProject.PROVIDERS in data:
            self.on_project_providers(data[mqtt.project.TopicProject.PROVIDERS])
        elif mqtt.project.TopicProject.SUBGINERIES in data:
            self.on_project_subgineries(data[mqtt.project.TopicProject.SUBGINERIES])
        elif mqtt.project.TopicProject.ENGINERIES in data:
            try:
                self.on_project_engineries(data[mqtt.project.TopicProject.ENGINERIES])
            except BaseException as ex:
                logging.exception(ex)
            finally:
                self.on_ready()

    def on_project_managers(self, data):
        for item in data:
            if item[ID] not in self._managers:
                self._managers[item[ID]] = {'data': item, PROVIDERS: {}}
        self.subscribe(mqtt.project.TopicProject.providers(self.id))

    def on_project_providers(self, data):
        for item in data:
            if item[MANAGER_ID] in self._managers:
                if item[ID] not in self._managers[item[MANAGER_ID]][PROVIDERS]:
                    manager = self._managers[item[MANAGER_ID]]
                    manager[PROVIDERS][item[ID]] = bam.Recipe(self.id, item[ID], item[TYPE], manager['data'])
        self.subscribe(mqtt.project.TopicProject.subgineries(self.id))

    def on_project_subgineries(self, data):
        allowed_types = [t.__name__ for t in self.subginery_types()]

        for item in data:
            try:
                if item.get('type', None) not in allowed_types:
                    continue

                self._subgineries[item[ID]] = bam.generator.generate_subginery(self.id, item)
                self._subgineries[item[ID]].set_adapter(self)
            except ProjectError as ex:
                logging.warning(ex)

        self.subscribe(mqtt.project.TopicProject.engineries(self.id))

    def on_project_engineries(self, data):
        allowed_types = [t.__name__ for t in self.enginery_types()]
        for item in data:
            if item.get('type', None) not in allowed_types:
                continue

            e_id = item[ID]
            if item.get(TYPE) == bam.engineries.LightingArea.__name__:
                continue

            if RECIPE in item and TYPE in item[RECIPE]:
                ingredients = sorted(item[RECIPE].get(INGREDIENTS, []), key=lambda ing: ing.get('index', -1))
                # ИО без привязки игнорируются
                # if not ingredients:
                #     continue
                try:
                    _eng = bam.generate_enginery(self.id, item)
                    _eng.set_adapter(self)
                    self._engineries[e_id] = _eng
                except ProjectError as ex:
                    logging.warning(ex)
                    continue
                else:
                    for subginery in self._subgineries.values():
                        if subginery.has_enginery(e_id):
                            _eng.set_subginery(subginery)
                            subginery.set_enginery(_eng)
                    for ingredient_data in ingredients:
                        for manager in self._managers.values():
                            recipe = manager[PROVIDERS].get(ingredient_data[ID])
                            if isinstance(recipe, bam.Recipe):
                                recipe.set_enginery(_eng)
                                _eng.set_ingredient(ingredient_data[TYPE], recipe)
                                break

    def on_ready(self):
        self.loaded = True
        logging.info('Project({}) loaded'.format(self.id))

        self.exploit_topics(self.subscribe)

    def exploit_topics(self, action):
        """Make action with topics"""

        for fut in self.STATE_TOPICS or ():
            action(mqtt.spread.topic.State(mqtt.spread.address.ProviderAddress(self.id, '+', '+', '+', '+', funit_type=fut)))

        for entity in self.enginery_types() + self.subginery_types():
            for funit in entity.settable_units or []:
                action(mqtt.spread.topic.Set(entity.get_address_broadcast(self.id, funit)))

            for funit in entity.gettable_units or []:
                action(mqtt.spread.topic.Renew(entity.get_address_broadcast(self.id, funit)))

    @log_command
    def on_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        try:
            self._on_command(topic, variable)
        except (EngineryNotPresent, ClassifierError, CommandError) as ex:
            # if variable.session_id:
            #     to_topic = mqtt.spread.topic.Reply(self.id, variable.session_id)
            #     self.publish(to_topic, dict(error=dict(code=123, message='{}'.format(ex))), False)
            logging.debug(ex)

    def _on_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        e_id = int(topic.entity_address.id)

        if isinstance(topic.entity_address, mqtt.spread.address.SubgineryAddress):
            if e_id in self._subgineries:
                self._subgineries[e_id].on_command(topic, variable)
            else:
                raise EngineryNotPresent(f'{topic.entity_address.type}({e_id}) is not present')
        elif isinstance(topic.entity_address, mqtt.spread.address.EngineryAddress):
            if e_id in self._engineries:
                self._engineries[e_id].on_command(topic, variable)
            else:
                raise EngineryNotPresent(f'{topic.entity_address.type}({e_id}) is not present')

    def on_state(self, topic: mqtt.spread.topic.State, variable: mqtt.spread.variable.Variable):
        m_id = int(topic.entity_address.manager_id)
        if m_id in self._managers:
            recipe = self._managers[m_id][PROVIDERS].get(topic.entity_address.id)
            if recipe:
                for eng in recipe.engs:
                    eng.on_state(topic, variable)
        else:
            raise ProviderNotPresent(topic.entity_address.type, topic.entity_address.id)

    def on_exit(self):
        super(Equipment, self).on_exit()

        for eng in self._engineries.values():
            try:
                eng.set_invalid()
            except BaseException as ex:
                logging.exception(ex)


class LightingEquipment(Equipment):
    EVENT_TOPICS = ('CurrentLuminosity', 'CurrentPresence', 'LastEvent')
    STATE_TOPICS = ('On', 'BrightnessLevel', 'Temperature', 'EngineryHoldTime') + EVENT_TOPICS

    @classmethod
    def enginery_types(cls): return [
        bam.engineries.DimmingLight, bam.engineries.SwitchingLight, bam.engineries.TunableWhiteLight,
        bam.engineries.PresenceSensor, bam.engineries.LightSensor,
        bam.engineries.LightingArea,
    ]

    @classmethod
    def subginery_types(cls): return [
        bam.subgineries.lighting.Lighting
    ]

    def __init__(self, *args, **kwargs):
        super(LightingEquipment, self).__init__(*args, **kwargs)

        self._lighting_areas = dict()
        self.button_presets = dict()

    def on_local(self, topic: mqtt.local.LAButtonPresets, data):
        if data:
            self.button_presets.setdefault(topic.entity_id, bam.engineries.lighting_area.ButtonPreset(topic.entity_id))
            self.button_presets[topic.entity_id].update(json.loads(data))
        else:
            self.button_presets.pop(topic.entity_id)

    def stack_for_reload(self):
        """Список коллекций объектов, для которых необходимо переопубликовать значения"""
        super(LightingEquipment, self).stack_for_reload()
        yield from self._lighting_areas

    def on_project_engineries(self, data):
        """Добавляет инженерные объекты в LA. Обязатегно выполнение после парсинга Simple ресайпов"""
        super(LightingEquipment, self).on_project_engineries(data)

        for item in data:
            if RECIPE in item and TYPE in item[RECIPE] and INGREDIENTS in item[RECIPE]:
                if item[TYPE] == bam.engineries.LightingArea.__name__:
                    la = bam.generate_enginery(self.id, item)
                    la.set_adapter(self)
                    for ingredient_data in sorted(item[RECIPE][INGREDIENTS], key=lambda ing: ing.get('index', -1)):
                        for manager in self._managers.values():
                            recipe = manager[PROVIDERS].get(ingredient_data[ID])
                            if recipe and isinstance(recipe, bam.Recipe):
                                recipe.set_la(la)
                                la.set_ingredient(ingredient_data[TYPE], recipe)
                                break
                    self._lighting_areas[item[ID]] = la

    def on_ready(self):
        for _la in self._lighting_areas.values():
            _la.get_info()

        super(LightingEquipment, self).on_ready()

    def exploit_topics(self, action):
        super(LightingEquipment, self).exploit_topics(action)
        action(mqtt.local.LAButtonPresets.broadcast(self.id))

    def _on_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        e_id = int(topic.entity_address.id)

        if topic.entity_address.type == bam.engineries.LightingArea.__name__:
            if e_id in self._lighting_areas:
                self._lighting_areas[e_id].on_command(topic, variable)
            else:
                raise EngineryNotPresent(f'{topic.entity_address.type}({e_id}) is not present')
        else:
            super(LightingEquipment, self)._on_command(topic, variable)

    def on_state(self, topic: mqtt.spread.topic.State, variable: mqtt.spread.variable.Variable):
        super(LightingEquipment, self).on_state(topic, variable)

        m_id = int(topic.entity_address.manager_id)
        recipe = self._managers[m_id][PROVIDERS].get(topic.entity_address.id)
        if recipe:
            for la in recipe.las:
                la.on_state(topic, variable)

    def on_event(self, topic: mqtt.spread.topic.State, variable: mqtt.spread.variable.Variable):
        if variable.invalid:
            return

        m_id = topic.entity_address.manager_id
        if m_id in self._managers:
            p_id = topic.entity_address.id
            for _la in self._lighting_areas.values():
                if p_id in _la.luminosities:
                    _la.on_luminosity_event(p_id, variable)

                elif p_id in _la.presences:
                    _la.on_presence_event(p_id, variable)

                elif p_id in _la.buttons:
                    _la.on_button_event(p_id, variable)
        else:
            raise ProviderNotPresent(topic.entity_address.type, topic.entity_address.entity_id)

    def on_exit(self):
        super(LightingEquipment, self).on_exit()

        for eng in self._lighting_areas.values():
            try:
                eng.set_invalid()
            except BaseException as ex:
                logging.exception(ex)


class ShadingEquipment(Equipment):
    STATE_TOPICS = ('PositionLevel', )

    @classmethod
    def enginery_types(cls): return [
        bam.engineries.shade.Shutter, bam.engineries.shade.Curtain, bam.engineries.shade.Blind,
    ]

    @classmethod
    def subginery_types(cls): return [
        bam.subgineries.shading.Shading
    ]
